/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: store.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"


/***********************\
*                       *
*  Storage management   *
*                       *
\***********************/

char *allocate(unsigned int size, unsigned int elt_size)
{
    char *p;
    if ((p = calloc (size, elt_size)) != NULL)
        return (p);
    if (store_verbosity)
      message("garbage collecting in allocate");
    collect_garbage ();
    while ((p = calloc (size, elt_size)) == NULL)
        error_break0 ("insufficient memory!");
    return (p);
}


char *strsave(char *s)                /* save string s somewhere */
{
    char   *p;
        
    p = allocate (strlen (s) + 1, sizeof (char));
    strcpy (p, s);
    return (p);
}

char *strapnd(char *s1, char *s2)            /* save append of s1 and s2 */
{
    char   *p;
        
    p = allocate (strlen (s1) + strlen (s2) + 1, sizeof (char));
    strcpy (p, s1);
    strcat (p, s2);
    return (p);
}

char *substr(char *s, int i, int j)         /* sub-string of s from i to j */
{
    int len = strlen(s);
    char   *p,
           *q;

    j = max (i - 1, min (j, len - 1));
    p = q = allocate ((unsigned)j - i + 2, sizeof (char));
    while (i <= j) {
        *p++ = *(s + i++);
    }
    *p = '\0';
    return (q);
}

void more_symbol_table(unsigned size)
{
    NODEPTR sp;
    if (store_verbosity)
        message ("extending symbol table");
    sp = nodep (allocate (size, sizeof (NODE)));
    SymbolTableTop = sp;
    SymbolTableMax = SymbolTableTop + size;
}

void more_heap(register unsigned size)
{
    HEAP h;
    register NODEPTR new_block;
    register unsigned i;
    if (store_verbosity)
        message("extending heap");
    h = heapp (allocate (1, sizeof (HEAPNODE)));
    h -> base = new_block = nodep (allocate (size, sizeof (NODE)));
    h -> top = h -> base + size;
    h -> next_block = Heap;
    for (i = 0; i < size - 1; i += 1)
        node_pnext(new_block[i]) = &(new_block[i+1]);
    node_pnext(new_block[i]) = NULL;
    FreeNodeList = new_block;
    Heap = h;
    HeapSize += size;
    FreeNodes += size;
}



int size_temp_node_list(void)
{
  int size = 0;
  NODEPTR np = TempNodeList;
  while (np != NULL) {
        size += 1;
        np = node_next(np);
  }
  return (size);
}

NODEPTR claim_node(void)
{
    register NODEPTR np;

    if (FreeNodeList == NULL) {
      if (store_verbosity)
        message("garbage collecting in claim_node");
      collect_garbage ();
      if (FreeNodeList == NULL) {
                more_heap (MAXHEAPSIZE);
      }
    }
    node_mark(np = FreeNodeList) = F;
    FreeNodeList = node_next(np);
    node_next(np) = TempNodeList;
    TempNodeList = np;
    TempNodes += 1;
    FreeNodes -= 1;
    return (np);
}

int free_heap_size(void)
{
    int     size = 0;
    NODEPTR np = FreeNodeList;
    while (np != NULL) {
        size += 1;
        np = node_next(np);
    }
    return (size);
}

NODEPTR node(unsigned type, NODEPTR v1, NODEPTR v2, NODEPTR v3, NODEPTR v4, NODEPTR v5)          /* Make a node with five children. */
{
    register NODEPTR np;
    np = claim_node();
    node_type(np) = type;
    node_lnk_1(np) = v1;
    node_lnk_2(np) = v2;
    node_lnk_3(np) = v3;
    node_lnk_4(np) = v4;
    node_lnk_5(np) = v5;
    return (np);
}

NODEPTR bool_node(unsigned int val)
{
    return (val ? True : False);
}

NODEPTR name_node (unsigned signal, char *name, unsigned int level, unsigned int off, unsigned framed)
{
    register NODEPTR np;

    np = claim_node();
    node_type(np) = signal ? STATE : STATIC;
    node_nm_name(np) = name;
    node_nm_level(np) = level;
    node_nm_call_level_offset(np) = 0;
    node_nm_off(np) = off;
    node_nm_framed(np) = framed;
    /*printf("temp_node_list_size=%d\n",size_temp_node_list());*/
    TempNodeList = node_next(np);
    node_next(np) = NULL;
    /*printf("temp_node_list_size=%d\n",size_temp_node_list());*/
    TempNodes -= 1;
    return(np);
}

NODEPTR loc_node(NODEPTR loc, unsigned is_next)
{
    register NODEPTR np = claim_node();
    node_type(np) = LOC;
    node_loc_ptr(np) = loc;
    node_loc_tonext(np) = is_next;
    return (np);
}

NODEPTR int_node(int64_t val)
{
    register NODEPTR ip = claim_node();
    node_type(ip) = INT;
    node_ival(ip) = val;
    return (ip);
}

NODEPTR string_node (char *val)
{
    register NODEPTR sp = claim_node();
    node_type(sp) = STRING;
    node_strval(sp) = val;
    return (sp);
}

NODEPTR float_node (double val)
{
    register NODEPTR fp = claim_node();
    node_type(fp) = FLOAT;
    node_flval(fp) = val;
    return (fp);
}

NODEPTR mpz_node (mpz_t val)
{
    register NODEPTR fp = claim_node();
    node_type(fp) = MPZ;
    mpz_init_set (node_mpzval(fp),  val);
    return (fp);
}

NODEPTR mpx_node (mpz_t val)
{
    register NODEPTR fp = claim_node();
    node_type(fp) = MPX;
    mpz_init_set (node_mpxval(fp),  val);
    return (fp);
}

NODEPTR mpz_node_str (char *val)
{
    register NODEPTR fp = claim_node();
    node_type(fp) = MPZ;
    mpz_init_set_str (node_mpzval(fp),  val, 10);
    /*gmp_printf("mpz_node.val == %Zd\n", node_mpzval(fp));*/
    return (fp);
}

NODEPTR mpx_node_str (char *val)
{
    register NODEPTR fp = claim_node();
    node_type(fp) = MPX;
    mpz_init_set_str (node_mpxval(fp),  val, 16);
    /*gmp_printf("mpz_node.val == %Zd\n", node_mpzval(fp));*/
    return (fp);
}

NODEPTR mpfr_node_str (char *val)
{
    register NODEPTR fp = claim_node();
    int ff;
    node_type(fp) = MPFR;
    mpfr_init_set_str (node_mpfrval(fp),  val, 10, MPFRRnd);
    /*printf("\nstring = %s valid = %d\n",val, ff);*/
    /*gmp_printf("mpz_node.val == %Zd\n", node_mpzval(fp));*/
    return (fp);
}


NODEPTR mpfr_node (mpfr_t val)
{
    register NODEPTR fp = claim_node();
    node_type(fp) = MPFR;
    mpfr_init_set(node_mpfrval(fp) , val, MPFRRnd);
    return (fp);
}


/*NODEPTR env_node (level,size,id)
unsigned level;
int size;
NODEPTR id;
{
    register NODEPTR ep;

    ep = claim_node();
    node_type(ep) = ENV;
    node_env_level(ep) = level;
    node_env_size(ep) = size;
    node_env_idlist(ep) = id;
    return(ep);
}
*/
/*
NODEPTR env_node1 (level,size,id,history)
unsigned level;
int size;
NODEPTR id;
unsigned int history;
{
    register NODEPTR ep;

    ep = claim_node();
    node_type(ep) = ENV;
    node_env_level(ep) = level;
    node_env_size(ep) = size;
    node_env_idlist(ep) = id;
    node_env_history(ep) = history;
    return(ep);
}
*/

NODEPTR env_node2 (unsigned level, int size, NODEPTR id, unsigned int history, unsigned framed)
{
    register NODEPTR ep;

    ep = claim_node();
    node_type(ep) = ENV;
    node_env_level(ep) = level;
    node_env_call_level_offset(ep) = 0;
    node_env_size(ep) = size;
    node_env_idlist(ep) = id;
    node_env_history(ep) = history;
    node_env_framed(ep) = framed ;
    return(ep);
}

NODEPTR bnd_node (NODEPTR env, NODEPTR blk, NODEPTR next)
{
    register NODEPTR bp;

    bp = claim_node();
    node_type(bp) = BND;
    node_bnd_env(bp) = env;
    node_bnd_blk(bp) = blk;
    node_bnd_next(bp) = next;
    return(bp);
}

NODEPTR list_node(NODEPTR blk, NODEPTR size_loc, NODEPTR base_loc)
{
    register NODEPTR lp;
       
    lp = claim_node();
    node_type(lp) = LIST;
    node_lst_blk(lp) = blk;
    node_lst_size(lp) = size_loc;
    node_lst_base(lp) = base_loc;
    return(lp);
}

NODEPTR block_node(unsigned int size, NODEPTR base_loc)
{
    register NODEPTR bp;

    bp = claim_node();
    node_type(bp) = MEMBLOCK;
    node_blk_size(bp) = size;
    node_blk_base(bp) = base_loc;
    return(bp);
}

unsigned int elist_size(NODEPTR elist)
{
    register unsigned int     size = 0;
    
    while (elist != NULL ) {
        size += 1;
        elist = node_lnk_2(elist);
    }
    return (size);
}

/* end of store.c */
