/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * fuzzy logic library
 */

load "../experimental/list_operators".


/********************** helper functions ************************/

define min(X,Y) = {
  if X <= Y then X else Y 
}.

define max(X,Y) = {
 if Y <= X then X else Y
}.


define plus = (% A,B : A + B ).

define minus = (% A,B : A - B ).

define mult = (% A,B : A * B ).

define sqr = (% A : A * A ).


define c0 = $0$.

define c1 = $1$.

define c2 = $2$.

define c3 = $3$.

define c4 = $4$.

define c10 = $10$. 

define min_list = (% L: foldl(min, c1, L)).
 
define max_list = (% L: foldl(max, c0, L)).

define sum_list = (% L: foldl(plus, c0, L)).


define lrange_fl =
  (% A,S,B : [A + (S*itof(i)) : i < 1+rint( (B - A) / S ) ] ).

define lapply =
  (% f,L : [ f(L[i]) : i<|L| ] ).

define list_of_maps =
  (% L, X: [ map(L[i], X) : i<|L| ] ).

/************* membership functions ************************/

define sig_mf =
  (% X0,X1 : (% X : c1/(c1+exp( -(X0 * (X-X1)) )) )).

define dsig_mf =
  (% X0,X1,X2,X3 : (% X :
    (c1/(c1+exp( -(X0 * (X-X1)) ))) -
    (c1/(c1+exp( -(X2 * (X-X3)) )))
   )
  ).

define psig_mf =
  (% X0,X1,X2,X3 : (% X :
    (c1/(c1+exp( -(X0 * (X-X1)) ))) *
    (c1/(c1+exp( -(X2 * (X-X3)) ))) 
   )
  ).

define gbell_mf =
 (% X0,X1,X2 : (% X : c1/(c1+fabs((X-X2)/X0)**(c2*X1)) )).

define gaussian_mf =
 (% X0,X1: (% X : exp( (-((X-X1)**c2)) / (c2*(X0**c2)) ) )).

define gaussian2_mf =
 (% X0,X1,X2,X3:
    (% X : if X <= X1 then exp( (-((X-X1)**c2)) / (c2 *(X0**c2)) )
           else if X <= X3 then c1
                else exp( (-((X-X3)**c2)) / (c2 *(X2**c2)) )    
    )
 ).

define triangle_mf =
 (% X0,X1,X2 : (% X : max(min((X-X0)/(X1-X0),(X2-X)/(X2-X1)), c0) ) ).

define trapezoid_mf =
 (% X0,X1,X2,X3 : (% X :
    max(min_list([(X-X0)/(X1-X0), c1, (X3-X)/(X3-X2)]), c0)
   )
 ).

define linz_mf =
 (% X0,X1 : (% X :
    if X0 < X1 then
     if X < X0 then c1 
     else if X <= X1 then (X1 - X) / (X1 - X0) else c0
    else (if X < X0 then c1 else c0)
   )
 ).   

define lins_mf =
 (% X0,X1 : (% X:
    if X0 < X1 then
     if X < X0 then c0
     else if X <= X1 then (X - X0) / (X1 - X0) else c1
    else (if X < X0 then c0 else c1)
   )
 ).  

define z_mf =
  (% X0,X1 : (% X :
     if X <= X0 then c1
     else if X <= (X0 + X1) / c2 then c1 - (c2 * (( (X-X0) / (X1-X0) )**c2) )
          else if X <= X1 then c2 * ( ( (X-X1) / (X1-X0) )**c2)
               else c0
    )
  ).  

define s_mf =
 (% X0,X1 : (% X :
    if X <= X0 then c0
    else if X <= (X0+X1)/c2 then c2 *( ( (X-X0) / (X1-X0) )**c2 )
          else if X <= X1 then c1 - (c2 * ( ( (X-X1) / (X1-X0) )**c2))
               else c1
   )
 ).

define pi_mf =
 (% X0,X1,X2,X3 : (% X :
    if X <= X0 then c0
    else if X <= (X0+X1) / c2 then c2 * ( ( (X-X0)/(X1-X0))**c2)
         else if X <= X1 then c1 - (c2 * ( ( (X-X1)/(X1-X0) )**c2))
              else if X <= X2 then c1
                   else if X <= (X2+X3)/c2 then c1 - (c2 *( ( (X-X2)/(X3-X2))**c2))
                        else if X <= X3 then c2 *( ( (X-X3)/(X3-X2))**c2)
                             else c0
   )
 ).  

/******************* fuzzy logic operators ***********************/

/* fuzzy AND */

define and_min = min.

define l_and_min =
  (% Xs,Ys: map_zip(and_min, Xs, Ys) ).

define and_min_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_and_min, head(L), tail(L) ) ).

define t_norm_min =
  (% mf1,mf2 : (% X : min(mf1(X), mf2(X)) ) ).


define and_product = mult.

define l_and_product =
  (% Xs,Ys: map_zip(and_product, Xs, Ys) ).

define and_product_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_and_product, head(L), tail(L) ) ).

define t_norm_product =
   (% mf1,mf2 : (% X :  mf1(X)*mf2(X) ) ).


define and_lukasiewicz =
 (% X,Y : max(c0, (X + Y) - c1) ).

define l_and_lukasiewicz =
  (% Xs,Ys: map_zip(and_lukasiewicz, Xs, Ys) ).

define and_lukasiewicz_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_and_lukasiewicz, head(L), tail(L) ) ).

define t_norm_lukasiewicz =
   (% mf1,mf2 : (% X : and_lukasiewicz(mf1(X) ,mf2(X)) ) ).


define and_drastic =
 (% X,Y : if max(X,Y) = c1 then min(X,Y) else c0 ).

define l_and_drastic =
  (% Xs,Ys: map_zip(and_drastic, Xs, Ys) ).

define and_drastice_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_and_drastic, head(L), tail(L) ) ).

define t_norm_drastic =
   (% mf1,mf2 : (% X : and_drastic(mf1(X), mf2(X)) )).


define and_nilpotent_min =
  (% X,Y : if X+Y > c1 then min(X,Y) else c0 ).

define l_and_nilpotent_min =
  (% Xs,Ys: map_zip(and_nilpotent_min, Xs, Ys) ).

define and_nilpotent_min_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_and_nilpotent_min, head(L), tail(L) ) ).

define t_norm_nilpotent_min =
   (% mf1,mf2 : (% X : and_nilpotent_min(mf1(X), mf2(X)) )).


define and_hamacher =
  (% X,Y : if X = c0 and Y = c0 then c0 else (X*Y) / ((X+Y) - (X*Y)) ).

define l_and_hamacher =
  (% Xs,Ys: map_zip(and_hamacher, Xs, Ys) ).

define and_hamacher_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_and_hamacher, head(L), tail(L) ) ).
  
define t_norm_hamacher =
  (% mf1,mf2 : (% X: and_hamacher(mf1(X), mf2(X)) )).
  

define and_einstein =
  (% X,Y: (X*Y) / (c2 - ((X+Y) - (X*Y))) ).

define l_and_einstein =
  (% Xs,Ys: map_zip(and_einstein, Xs, Ys) ).

define and_einstein_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_and_einstein, head(L), tail(L) ) ).

define t_norm_einstein =
  (% mf1,mf2 : (% X: and_einstein(mf1(X), mf2(X)) ) ). 


/* fuzzy OR */

define or_max = max. 

define l_or_max =
  (% Xs,Ys : map_zip(or_max, Xs, Ys) ).

define or_max_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_or_max, head(L), tail(L) ) ).

define t_conorm_max =
  (% mf1,mf2: (% X: max(mf1(X),mf2(X)) ) ).
  

define or_sum =
  (% X,Y : ((X+Y) - (X*Y)) ).

define l_or_sum =
  (% Xs, Ys : map_zip(or_sum, Xs, Ys)).

define or_sum_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_or_sum, head(L), tail(L) ) ).

define t_conorm_sum =
  (% mf1,mf2: (% X : or_sum(mf1(X), mf2(X)) )).
  

define or_lukasiewicz =
  (% X,Y : min(X+Y,c1) ).

define l_or_lukasiewicz =
  (% Xs, Ys : map_zip(or_lukasiewicz, Xs, Ys)).

define or_lukasiewicz_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_or_lukasiewicz, head(L), tail(L) ) ).

define t_conorm_lukasiewicz =
  (% mf1,mf2 : (% X: or_lukasiewicz(mf1(X), mf2(X)) ) ).


define or_drastic =
  (% X,Y :  if min(X,Y) = c0 then max(X,Y) else c1 ).

define l_or_drastic =
  (% Xs, Ys : map_zip(or_drastic, Xs, Ys)).

define or_drastic_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_or_drastic, head(L), tail(L) ) ).

define t_conorm_drastic =
  (% mf1,mf2 : (% X : or_drastic(mf1(X), mf2(X)) ) ).
  
      
define or_nilpotent_max =
  (% X,Y : if X+Y < c1 then max(X,Y) else c1 ). 

define l_or_nilpotent_max =
  (% Xs, Ys : map_zip(or_nilpotent_max, Xs, Ys)).

define or_nilpotent_max_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_or_nilpotent_max, head(L), tail(L) ) ).

define t_conorm_nilpotent_max =
   (% mf1,mf2: (% X : or_nilpotent_max(mf1(X), mf2(X)) ) ).
   

define or_hamacher =
 (% X,Y : if X = c1 and Y = c1 then c1 else ((X + Y) - (c2*X*Y)) / (c1 - (X * Y)) ).

define l_or_hamacher =
  (% Xs, Ys : map_zip(or_hamacher, Xs, Ys)).

define or_hamacher_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_or_hamacher, head(L), tail(L) ) ).

define t_conorm_hamacher =
  (% mf1,mf2 : (% X : or_hamacher(mf1(X), mf2(X)) ) ).
  

define or_einstein =
  (% X,Y : (X+Y)/(c1 + (X*Y)) ).

define l_or_einstein =
  (% Xs, Ys : map_zip(or_einstein, Xs, Ys)).

define or_einstein_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_or_einstein, head(L), tail(L) ) ).

define t_conorm_einstein =
   (% mf1,mf2 : (% X : or_einstein(mf1(X), mf2(X)) ) ).
   

/* fuzzy NOT */

define f_not = (% X : c1 - X ).

define l_f_not = (% Xs: map(f_not, Xs) ).
 
define negator = (% mf : (% X : c1 - mf(X) ) ).

/*    */

define b_difference =
 (% X,Y : max(c0, (X+Y) - c1) ).

define l_b_difference =
  (% Xs, Ys : map_zip(b_difference, Xs, Ys)).

define b_difference_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_b_difference, head(L), tail(L) ) ).

define bounded_difference =
   (% mf1, mf2 : (% X : b_difference(mf1(X), mf2(X)) ) ).


define b_sum =
  (% X,Y : min(c1, X + Y) ).

define l_b_sum =
  (% Xs, Ys : map_zip(b_sum, Xs, Ys)).

define b_sum_list =
  (% L : if |L| <= 1 then head(L) else foldl(l_b_sum, head(L), tail(L) ) ).

define bounded_sum =
   (% mf1,mf2 : (% X: b_sum(mf1(X), mf2(X)) ) ). 



/* fuzzy IMPLICATION */

define imp_larsen = mult.

define l_imp_larsen =
  (% Xs, Ys : map_zip(imp_larsen, Xs, Ys)).

define residuum_larsen =
  (% mf1,mf2 : (% X: imp_larsen(mf1(X), mf2(X)) ) ).


define imp_mamdani = min.

define l_imp_mamdani =
  (% Xs, Ys : map_zip(imp_mamdani, Xs, Ys)).

define residuum_mamdani =
  (% mf1,mf2 : (% X: imp_mamdani(mf1(X), mf2(X)) ) ).


/* goedel */
define imp_min =
  (% X,Y : if X <= Y then c1 else Y ).

define l_imp_min =
  (% Xs, Ys : map_zip(imp_min, Xs, Ys)).

define residuum_min =
  (% mf1,mf2 : (% X: imp_min(mf1(X), mf2(X)) ) ).


/* gaines */
define imp_goguen =
  (% X,Y : if X <= Y then c1 else Y / X ).

define l_imp_goguen =
  (% Xs, Ys : map_zip(imp_goguen, Xs, Ys)).

define residuum_goguen =
  (% mf1,mf2 : (% X : imp_goguen(mf1(X), mf2(X)) ) ).


define imp_lukasiewicz =
  (% X,Y : if X <= Y then c1 else ((c1 - X) + Y ) ). 

define l_imp_lukasiewicz =
  (% Xs, Ys : map_zip(imp_lukasiewicz, Xs, Ys)).

define residuum_lukasiewicz =
  (% mf1,mf2 : (% X : imp_lukasiewicz(mf1(X), mf2(X)) ) ).


/* kleene-dienes */
define imp_nilpotent_min =
  (% X,Y : if X <= Y then c1 else max(c1 - X, Y) ).

define l_imp_nilpotent_min  =
  (% Xs, Ys : map_zip(imp_nilpotent_min, Xs, Ys)).

define residuum_nilpotent_min =
  (% mf1,mf2 : (% X: imp_nilpotent_min(mf1(X), mf2(X)) ) ).


define imp_standard_strict =
  (% X,Y : if X <= Y then c1 else c0 ).

define l_imp_standard_strict  =
  (% Xs, Ys : map_zip(imp_standard_strict, Xs, Ys)).

define residuum_standard_strict =
  (% mf1,mf2 : (% X: imp_standard_strict(mf1(X), mf2(X)) ) ).


/* kleene-dienes-lukasiewicz */
define imp_dkl =
  (% X,Y : ((c1 - X) + X*Y)).

define l_imp_dkl  =
  (% Xs, Ys : map_zip(imp_dkl, Xs, Ys)).

define residuum_dkl =
  (% mf1,mf2 : (% X: imp_dkl(mf1(X), mf2(X)) ) ).



/***************** hedges ****************************************/

define somewhat =
  (% X : sqrt(X) ).

define l_somewhat =
  (% Xs: map(somewhat, Xs) ).

define hedge_somewhat =
  (% mf : (% X : somewhat(mf(X)) )).

define very =
  (% X : sqr(X) ).

define l_very =
  (% Xs: map(very, Xs) ).

define hedge_very =
  (% mf : (% X : very(mf(X)) ) ).

define extremely =
 (% X : X**c3 ).

define l_extremely =
  (% Xs: map(extremely, Xs) ).

define hedge_extremely =
  (% mf : (% X : extremely(mf(X)) ) ).

define very_very =
  (% X : X**c4 ).

define l_very_very =
  (% Xs: map(very_very, Xs) ).

define hedge_very_very =
  (% mf : (% X : very_very(mf(X)) ) ).


define h_custom =
  (% dd : (% X : X**(itof(dd) / c10) ) ).

define l_h_custom =
  (% dd : (% Xs: map(h_custom(dd), Xs) ) ).

define hedge_custom =
 (% mf : (% dd : (% X : h_custom(dd)(mf(X)) ) ) ).
 

/***************** defuzzification *******************************/

/*
 * \int_a^b f(x) dx = 
 * 1/2 * \sum_{i=0}^N (x_{i+1} - x_i)*(f(x_i) + f(x_{i+1}))
 * where x_i < x_{i+1}, i \in [0..N],
 * x_0 = a and x_{N+1} = b
 */

define list_minus = 
  (% X,Y : map_zip( minus, X, Y) ).

define list_plus =
  (% X,Y : map_zip( plus, X, Y) ).

define list_mult =
  (% X,Y : map_zip( mult, X, Y ) ).

define trapz_a =
  (% X : list_minus(tail(X), X) ).

define trapz_b =
  (% X : list_plus(X, tail(X)) ).

define trapz_c =
  (% X, Y: list_mult(trapz_a(X), trapz_b(Y)) ).

define trapz =
  (% X,Y : (c1 / c2) * sum_list( trapz_c(X,Y) ) ). 

define centroid =
  (% X, Y : trapz(X, list_mult(X, Y) )  / trapz (X, Y) ).

/*
* X[(the n:  trapz(take(n,X),take(n,Y)) = trapz(drop(n,X),drop(n,Y)) )]
*/
 define binsearch(X, Y, L, R, M, H) = {
    if L > R then M 
    else 
      /*if trapz( (take(M, X)), (take(M, Y)) ) = H
      then M
      else*/ if trapz( X[0..M], Y[0..M] ) < H
           then binsearch(X, Y, M+1, R, (M+1)+((R-(M+1))/2), H)
           else binsearch(X, Y, L, M-1, L+(((M-1)-L)/2), H)
 }. 
     
 
define bisector =
  (% X,Y :  X[binsearch(X, Y, 0, |X|-1, (|X|-1)/2, trapz(X,Y) / $2$)]  ). 


define list_mean =
  (% X: sum_list(X) / itof(|X|)).

  
define mom =
 (% X, Y, M:
   list_mean (
      (map ((% p : p[0] ), 
            (filter ((% z : z[1] = M),
                     ( zip(X,Y) )
                    ) 
            )
           ) 
      )    
   )
 ).  


define som =
 (% X,Y,M :
  head (
         (map ((% d: d[0]), 
               (filter ((% z : z[1] = M),
                        ( zip(X,Y) )
                       ) 
               )
              )
         )
        )      
 ).

define lom =
 (% X,Y,M :
    last(
         (map ((% d: d[0]),  
               (filter ((% z : z[1] = M),
                        ( zip(X,Y) )
                       ) 
               )
              ) 
         )
     )   
 ).

define wtaver =
  (% X,Y : sum_list(list_mult(X,Y)) / sum_list(Y) ).

define wtsum =
  (% X,Y : sum_list(list_mult(X,Y)) ).



